/*
** Module   : DBUtils.java
** Abstract : Static helper methods for DB services
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060712   @28043 Created initial version. Static utility methods for the persistence
**                           framework.
** 002 ECF 20060714   @28136 Made class and most methods public. Access is required from the
**                           schema package for import.
** 003 ECF 20070416   @33031 Added handleInterrupt(). Inspects an exception and throws
**                           StopConditionException if it finds InterruptedException within the
**                           exception's causal chain.
** 004 EVL 20070609   @34005 Adding explicit import of the class org.hibernate.type.Type to
**                           eliminate conflict with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 005 ECF 20070705   @34358 Enhanced makeTypeArray(). Entity parameters are now allowed.
** 006 ECF 20071207   @36260 Changed findDMOGetterMethod() signature. Now
**                           throws PersistenceException rather than IllegalArgumentException.
** 007 ECF 20080310   @37469 Added new database dialect support. Method composePropertyName()
**                           now supports computed columns. Also added new variant of method
**                           getDMOPropertyType().
** 008 ECF 20080318   @37542 Fixed makeTypeArray(). It was not properly handling unresolved
**                           (i.e. Resolvable) query substitution parameters.
** 009 ECF 20080330   @37731 Fixed composePropertyName() (package private variant). It was
**                           qualifying property name with an alias at the wrong point.
** 010 ECF 20080510   @38483 Added new methods. entityForDMOInterface() and dmoClassForEntity().
** 011 ECF 20080806   @39318 Added rowid support. New mappings for 64-bit integers. Import
**                           persist.type package, since UserType classes moved.
** 012 CA  20080815   @39454 Changed handleInterrupt to handleException.
**                           The method will handle database connection errors beside interrupt.
** 013 GES 20090424   @41936 Import change.
** 014 SVL 20090225   @41365 Into TYPES map added mapping for recid.
** 015 ECF 20090525   @42972 Added composePropertyName() variant. Added dirty parameter to
**                           differentiate between primary and dirty database dialects when
**                           composing property names.
** 016 ECF 20090717   @43228 Created workaround for remote dirty database lookup failure.
**                           We drop back to using primary database's dialect in
**                           composePropertyName().
** 017 ECF 20090728   @43416 Fixed getDMOPropertyType(). RecordBuffer change
**                           had regressed this method.
** 018 ECF 20090817   @43682 Added new *Field types to TYPES map.
** 019 VMN 20130416          Added method getGetterSetterType().
** 020 SVL 20130331          Upgraded to Hibernate 4.
** 021 OM  20130430          Added new int64 and Int64Field to TYPES map.
** 022 OM  20130527          Remapped DatetimeTzUserType as CompositeCustomType.
**                           For the moment this is disabled (1584#150).
** 023 SVL 20130716          Use system class loader in dmoClassForEntity.
** 024 ECF 20131010          Added metadata support.
** 025 OM  20140410          Fixed support case-sensitive fields in indexes.
** 026 VMN 20140427          Added enhance schema name conversion support for hint "escape"
**                           attribute.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.cfg.*;
import org.hibernate.type.*;
import org.hibernate.type.Type;
import com.goldencode.util.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.type.*;
import com.goldencode.p2j.util.*;

/**
 * A collection useful static methods used to support the persistence
 * framework.
 */
public final class DBUtils
{
   /** Logger */
   private static final Log LOG = Utils.getLog(DBUtils.class);
   
   /** Mapping of wrapper data types to Hibernate <code>Type</code>s */
   private static final Map<Class<?>, Type> TYPES = new HashMap<Class<?>, Type>();
   
   // Initialize the TYPES map.
   static
   {
      TYPES.put(integer.class   , new CustomType(new IntegerUserType()));
      TYPES.put(recid.class     , new CustomType(new IntegerUserType()));
      TYPES.put(decimal.class   , new CustomType(new DecimalUserType()));
      TYPES.put(character.class , new CustomType(new CharacterUserType()));
      TYPES.put(logical.class   , new CustomType(new LogicalUserType()));
      TYPES.put(date.class      , new CustomType(new DateUserType()));
      TYPES.put(raw.class       , new CustomType(new RawUserType()));
      TYPES.put(rowid.class     , new CustomType(new RowIDUserType()));
      TYPES.put(int64.class     , new CustomType(new Int64UserType()));
      TYPES.put(datetime.class  , new CustomType(new DatetimeUserType()));
      // TYPES.put(datetimetz.class, new CompositeCustomType(new DatetimeTzUserType()));
      TYPES.put(datetimetz.class, new CustomType(new DatetimeUserType()));
      TYPES.put(Integer.class   , StandardBasicTypes.INTEGER);
      TYPES.put(Long.class      , StandardBasicTypes.LONG);
      TYPES.put(IntegerField.class   , new CustomType(new IntegerUserType()));
      TYPES.put(DecimalField.class   , new CustomType(new DecimalUserType()));
      TYPES.put(CharacterField.class , new CustomType(new CharacterUserType()));
      TYPES.put(LogicalField.class   , new CustomType(new LogicalUserType()));
      TYPES.put(DateField.class      , new CustomType(new DateUserType()));
      TYPES.put(RawField.class       , new CustomType(new RawUserType()));
      TYPES.put(Int64Field.class     , new CustomType(new Int64UserType()));
      TYPES.put(DatetimeField.class  , new CustomType(new DatetimeUserType()));
      //TYPES.put(DatetimeTzField.class, new CompositeCustomType(new DatetimeTzUserType()));
      TYPES.put(DatetimeTzField.class, new CustomType(new DatetimeUserType()));
   }
   
   /**
    * Get the mapping of Java class to Hibernate data type used by P2J.
    * 
    * @return  Unmodifiable map of classes to Hibernate types.
    */
   public static Map<Class<?>, Type> getHibernateTypeMap()
   {
      return Collections.unmodifiableMap(TYPES);
   }
   
   /**
    * Convenience method to create an array of Hibernate <code>Type</code>
    * objects to match the given array of substitution parameters.  It is
    * assumed that neither <code>args</code> nor any of its elements are
    * <code>null</code>.
    *
    * @param   args
    *          Array of HQL query substitution parameters.
    *
    * @return  Array of matching, Hibernate <code>Type</code> instances.
    * 
    * @throws  IllegalArgumentException
    *          if <code>args</code> contains a parameter of unknown type.
    */
   public static Type[] makeTypeArray(Object[] args)
   {
      return makeTypeArray(args, new Type[args.length]);
   }
   
   /**
    * Convenience method to populate an array of Hibernate <code>Type</code>
    * objects to match the given array of substitution parameters.  It is
    * assumed that neither <code>args</code> nor any of its elements are
    * <code>null</code>.
    * <p>
    * This implementation handles both resolved (<code>BaseDataType</code>)
    * and unresolved (<code>Resolvable</code>) arguments.  The latter is
    * possible because we only care about the data type, not the actual value,
    * of the parameters.
    * <p>
    * If any element in the <code>types</code> array is non-<code>null</code>,
    * it is skipped.
    *
    * @param   args
    *          Array of HQL query substitution parameters.
    * @param   types
    *          Array of the same size as <code>args</code>;  some elements
    *          may be pre-populated.
    *
    * @return  Array of matching, Hibernate <code>Type</code> instances.
    * 
    * @throws  IllegalArgumentException
    *          if <code>args</code> contains a parameter of unknown type.
    */
   public static Type[] makeTypeArray(Object[] args, Type[] types)
   {
      int len = args.length;
      for (int i = 0; i < len; i++)
      {
         if (types[i] != null)
         {
            // Type already known;  don't replace it.
            continue;
         }
         
         Object next = args[i];
         Class<?> cls = (next instanceof Resolvable
                         ? ((Resolvable) next).getType()
                         : next.getClass());
         
         // Look up the Hibernate type associated with the Java type.
         Type type = TYPES.get(cls);
         
         // If no type matched, assume we are using an entity parameter.
         if (type == null)
         {
            // This will throw IllegalArgumentException if class cannot be
            // found among mapped, persistent classes.
            Class normalClass = DatabaseManager.normalizeDMOClass(cls);
            String schema = DatabaseManager.getSchemaByClass(normalClass);
            Configuration cfg = DatabaseManager.getConfigBySchema(schema);
            type = cfg.getTypeResolver().getTypeFactory().manyToOne(normalClass.getName());
         }

         types[i] = type;
      }
      
      return types;
   }
   
   /**
    * Report the data type of a DMO property. Reflection is used to look up
    * the field which backs the specified property. This will not work for
    * list/composite properties, since these are not backed by a simple field.
    * In these cases, the indexed getter method is introspected and the data
    * type which it returns is reported.
    * <p>
    * Note:  this method will only work while the {@link RecordBuffer}
    * represented by the variable name <code>dmoAlias</code> has had a scope
    * opened upon it by business logic, within the current context.
    *
    * @param   dmoAlias
    *          Variable
    * @param   name
    *          Unqualified DMO property name.
    *
    * @return  Class which indicates the data type of the specified property.
    *
    * @throws  NullPointerException
    *          if <code>dmoAlias</code> does not represent a record buffer
    *          whose scope is currently open.
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static Class<?> getDMOPropertyType(String dmoAlias, String name)
   throws PersistenceException
   {
      RecordBuffer buffer = BufferManager.get().lookupBuffer(dmoAlias);
      if (buffer == null)
      {
         throw new NullPointerException(
            "No record buffer found for alias '" + dmoAlias + "'");
      }
      buffer.initialize();
      
      return getDMOPropertyType(buffer.getDMOImplementationClass(), name);
   }
   
   /**
    * Report the data type of a DMO property. Reflection is used to look up
    * the field which backs the specified property. This will not work for
    * list/composite properties, since these are not backed by a simple field.
    * In these cases, the indexed getter method is introspected and the data
    * type which it returns is reported.
    *
    * @param   dmoClass
    *          DMO implementation class.
    * @param   name
    *          Unqualified DMO property name.
    *
    * @return  Class which indicates the data type of the specified property.
    *
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static Class<?> getDMOPropertyType(Class<?> dmoClass, String name)
   throws PersistenceException
   {
      try
      {
         Class<?> propClass = null;
         
         // The easy way (by field)...
         try
         {
            propClass = dmoClass.getDeclaredField(name).getType();
         }
         catch (NoSuchFieldException nsfExc)
         {
         }
         
         // ...or the hard way (by indexed getter method).
         if (propClass == null || List.class.isAssignableFrom(propClass))
         {
            Class<?>[] parmTypes = new Class<?>[] { Integer.TYPE };
            Method getter = findDMOGetterMethod(dmoClass, name, parmTypes);
            propClass = getter.getReturnType();
         }
         
         return propClass;
      }
      catch (Exception exc)
      {
         String msg = "Error detecting DMO field type:  " +
                      dmoClass.getName() + "." + name;
                      
         if (LOG.isErrorEnabled())
         {
            LOG.error(msg, exc);
         }
         
         throw new PersistenceException(msg, exc);
      }
   }
   
   /**
    * Compose a DMO property name within the given string buffer. For
    * non-text properties, this will simply emit the property name, qualified
    * by the proper DMO alias.  For text properties, however, the qualified
    * property name is enclosed within a SQL function which trims trailing
    * whitespace.  For case-insensitive text properties, this expression is
    * enclosed by a SQL function which uppercases the trimmed result string.
    * For a text property whose qualified name is <code>foo.bar</code>, the
    * end result would be:
    * <pre>
    *    upper(rtrim(foo.bar, ' \t\n\r'))
    * </pre>
    * For a non-text property <code>foo.bar</code>, <code>foo.bar</code> is
    * emitted.
    *
    * @param   isCharacter
    *          <code>true</code> for text properties;  <code>false</code> for
    *          non-text properties.
    * @param   ignoreCase
    *          <code>true</code> if the name expression must be wrapped in an
    *          uppercasing function;  <code>false</code> for case-sensitive
    *          text properties, as well as for non-text properties.
    * @param   computedColumnPrefix
    *          If a computed column is needed instead of embedding the DMO
    *          property name within a function, this parameter should be the
    *          non-null prefix to be prepended before the property name.
    *          If <code>null</code>, it is assumed function use is acceptable.
    * @param   alias
    *          Optional alias used to qualify <code>name</code>.  Ignored if
    *          <code>null</code>.
    * @param   name
    *          Unqualified DMO property name.
    * @param   buf
    *          String buffer into which results are emitted.
    */
   public static void composePropertyName(boolean isCharacter,
                                          boolean ignoreCase,
                                          String computedColumnPrefix,
                                          String alias,
                                          String name,
                                          StringBuilder buf)
   {
      String qualifier = "";
      if (alias != null)
      {
         StringBuilder sb = new StringBuilder(alias);
         sb.append('.');
         qualifier = sb.toString();
      }
      
      if (!isCharacter)
      {
         buf.append(qualifier).append(name);
         
         return;
      }
      
      if (computedColumnPrefix != null)
      {
         buf.append(qualifier);
         buf.append(PropertyHelper.getPrefixedParameter(name, computedColumnPrefix));

         return;
      }
      
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      
      buf.append("rtrim(");
      buf.append(qualifier);
      buf.append(name);
      buf.append(", ' \t\n\r')");
      
      if (ignoreCase)
      {
         buf.append(")");
      }
   }
   
   /**
    * Get the DMO implementation class for the given DMO entity name.
    * 
    * @param   entityName
    *          DMO entity name.
    * 
    * @return  DMO implementation class.
    * 
    * @throws  IllegalArgumentException
    *          if <code>entityName</code> does not represent a loaded class,
    *          or if there is any other error loading the class.
    */
   public static Class<?> dmoClassForEntity(String entityName)
   {
      try
      {
         return Class.forName(entityName, true, ClassLoader.getSystemClassLoader());
      }
      catch (Exception exc)
      {
         throw new IllegalArgumentException(
            "Invalid DMO entity name:  " + entityName,
            exc);
      }
   }
   
   /**
    * Generate the DMO implementation class name (the entity name), given a
    * DMO interface.  This codifies the convention of adding the text "Impl"
    * to the end of the interface name to come up with the implementation
    * name.
    * 
    * @param   dmoIface
    *          DMO interface.
    * 
    * @return  DMO implementation class name.
    */
   public static String entityForDMOInterface(Class<?> dmoIface)
   {
      StringBuilder buf = new StringBuilder(dmoIface.getName());
      buf.append("Impl");
      
      return buf.toString();
   }
   
   /**
    * Handle an <code>InterruptedException</code>, which may be wrapped in an
    * exception causal chain, by throwing <code>StopConditionException</code>.
    * Alternatively, if the error represents a database connection problem,
    * <code>DatabaseConnectionException</code> is thrown.  Determination of
    * the latter type of problem is dialect-specific.
    * <p>
    * If neither of these conditions is found to be the cause of the error,
    * this method returns normally.
    *
    * @param   database
    *          The database in use when the exception was thrown. May be
    *          <code>null</code> if no database is used.
    * @param   exc
    *          An exception to inspect.
    *
    * @throws  StopConditionException
    *          if <code>exc</code> or any exception in its causal chain is an
    *          instance of <code>InterruptException</code>.
    * @throws  DatabaseConnectionException
    *          if <code>exc</code> or any exception in its causal chain is
    *          determined to be a database connection error.
    */
   public static void handleException(Database database, Exception exc)
   {
      if (database != null)
      {
         P2JDialect dialect = null;
         dialect = (P2JDialect) DatabaseManager.getDialect(database);
         
         if (dialect.isConnectionError(exc))
         {
            ErrorManager.displayAbend(1006, 
               "Database " + database.getName() + " not connected");
            ErrorManager.displayError(492, 
               "Unable to run startup procedure");
            throw new DatabaseConnectionException(exc, database);
         }
      }

      Throwable next = exc;
      while (next != null)
      {
         if (next instanceof InterruptedException)
         {
            throw new StopConditionException(exc);
         }
         next = next.getCause();
      }
   }

   /**
    * This method returns return type or parameter type depending on getter/setter method name
    * prefix.
    *
    * @param   method
    *          Method.
    *
    * @return  Return type or parameter type depending on getter/setter method name prefix.
    */
   static Class<?> getGetterSetterType(Method method)
   {
      String methodName = method.getName();
      if (methodName.startsWith("set"))
      {
         Class<?>[] parameterTypes = method.getParameterTypes();

         // if indexed: first parameter is array index, second one is value
         return parameterTypes[parameterTypes.length - 1];
      }
      else if (methodName.startsWith("get") ||  methodName.startsWith("is"))
      {
         return method.getReturnType();
      }
      else
      {
         throw new UnsupportedOperationException(
            "Method getGetterSetterType is acceptable only for getters/setters");
      }
   }

   /**
    * Look up the getter method associated with the given property.  Tests
    * for both <code>get</code> and <code>is</code> variants.
    *
    * @param   dmoClass
    *          DMO implementation class.
    * @param   propName
    *          Unqualified DMO property name.
    * @param   parmTypes
    *          Array of data types which are expected by the getter method.
    *          This will generally be <code>null</code> (or an empty array),
    *          or an array with one element of type <code>Integer.TYPE</code>
    *          (representing the index value passed to an indexed getter
    *          method).
    *
    * @return  Getter method for the given property.
    *
    * @throws  PersistenceException
    *          if <code>propName</code> does not represent a property of
    *          <code>dmoClass</code>.
    */
   static Method findDMOGetterMethod(Class<?> dmoClass, String propName, Class<?>[] parmTypes)
   throws PersistenceException
   {
      String methName;
      
      try
      {
         // First try with a "get" prefix (the majority of cases).
         methName = StringHelper.generateBeanName(propName, true, false);
         
         return dmoClass.getDeclaredMethod(methName, parmTypes);
      }
      catch (NoSuchMethodException exc)
      {
      }
      
      try
      {
         // As a fallback, try with an "is" prefix, in case property is
         // a logical.
         methName = StringHelper.generateBeanName(propName, true, true);
         
         return dmoClass.getDeclaredMethod(methName, parmTypes);
      }
      catch (NoSuchMethodException exc2)
      {
         throw new PersistenceException(
            dmoClass.getName() + " does not contain property '" + propName +
            "'");
      }
   }
   
   /**
    * Compose a DMO property name within the given string buffer. For
    * non-text properties, this will simply emit the property name, qualified
    * by the proper DMO alias.  For text properties, however, the qualified
    * property name is enclosed within a SQL function which trims trailing
    * whitespace.  For case-insensitive text properties, this expression is
    * enclosed by a SQL function which uppercases the trimmed result string.
    * For a text property whose qualified name is <code>foo.bar</code>, the
    * end result would be:
    * <pre>
    *    upper(rtrim(foo.bar, ' \t\n\r'))
    * </pre>
    * <p>
    * For a non-text property <code>foo.bar</code>, <code>foo.bar</code> is
    * emitted.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO's alias,
    *          interface, and implementation class.
    * @param   name
    *          Unqualified DMO property name.
    * @param   buf
    *          String buffer into which results are emitted.
    *
    * @throws  PersistenceException
    *          if any errors occur looking up metadata regarding the property.
    */
   static void composePropertyName(RecordBuffer buffer, String name, StringBuilder buf)
   throws PersistenceException
   {
      composePropertyName(buffer, name, buf, Database.Type.PRIMARY);
   }
   
   /**
    * Compose a DMO property name within the given string buffer. For
    * non-text properties, this will simply emit the property name, qualified
    * by the proper DMO alias.  For text properties, however, the qualified
    * property name may be enclosed within a SQL function which trims trailing
    * whitespace.  For case-insensitive text properties, this expression is
    * enclosed by a SQL function which uppercases the trimmed result string.
    * For a text property whose qualified name is <code>foo.bar</code>, the
    * end result would be:
    * <pre>
    *    upper(rtrim(foo.bar, ' \t\n\r'))
    * </pre>
    * <p>
    * In cases where a special computed column name is required instead of the
    * enclosing function syntax noted above, such a name is generated instead.
    * <p>
    * For a non-text property <code>foo.bar</code>, <code>foo.bar</code> is
    * emitted.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO's alias,
    *          interface, and implementation class.
    * @param   name
    *          Unqualified DMO property name.
    * @param   buf
    *          String buffer into which results are emitted.
    * @param   type
    *          Database type, which drives the dialect to use to compose the name.
    * 
    * @throws  PersistenceException
    *          if any errors occur looking up metadata regarding the property.
    */
   static void composePropertyName(RecordBuffer buffer,
                                   String name,
                                   StringBuilder buf,
                                   Database.Type type)
   throws PersistenceException
   {
      boolean isCharacter = false;
      boolean ignoreCase = false;
      String computedColumnPrefix = null;
      
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      if (character.class.equals(getDMOPropertyType(dmoClass, name)))
      {
         String schema = buffer.getSchema();
         Class<?> dmoIface = buffer.getDMOInterface();
         isCharacter = true;
         ignoreCase = !DMOIndex.isCaseSensitive(schema, dmoIface, name);
      }
      
      if (isCharacter)
      {
         P2JDialect dialect = null;
         
         if (!Database.Type.PRIMARY.equals(type))
         {
            Database secDB = new Database(buffer.getDatabase(), type);
            dialect = (P2JDialect) DatabaseManager.getDialect(secDB);
         }
         
         if (dialect == null)
         {
            dialect = buffer.getDialect();
         }
            
         if (dialect.needsComputedColumns())
         {
            Boolean ccIgnoreCase = DatabaseManager.getIgnoreCase(dmoClass, name);
            if (ccIgnoreCase != null)
            {
               computedColumnPrefix = dialect.getComputedColumnPrefix(!ignoreCase);
            }
         }
      }
      
      composePropertyName(isCharacter,
                          ignoreCase,
                          computedColumnPrefix,
                          buffer.getDMOAlias(),
                          name,
                          buf);
   }
}
